from itertools import accumulate, combinations, product
import inspect
from artiruno.preorder import PreorderedSet, IC, LT, EQ, GT
from artiruno.util import cmp, choose2

class Jump(Exception):
    def __init__(self, value):
        self.value = value

class Abort(Exception): pass

def vda(*args, **kwargs):
    import asyncio
    return asyncio.run(avda(*args, **kwargs))

async def avda(
        criteria, alts = None, asker = None, find_best = None,
        max_dev = 2, allowed_pairs_callback = lambda x: None):
    """
- `criteria` is an iterable of iterables specifying the levels of
  each criterion. Within a criterion, we assume that later levels are
  better.
- `alts` is a list of the alternatives, each represented as a tuple of
  the criteria. If it's `None`, we use the entire item space.
- `asker` should be a callable object f(a, b) that returns
    GT (1) if a is better
    LT (-1) if b is better
    0 (EQ) if they're equally good
- `find_best` can be set to an integer. If so, VDA will aim to
  identify the top `find_best` items and stop there. Otherwise, we'll
  try to compare all the alternatives.
- `max_dev` sets the maximum number of criteria on which hypothetical
  items can deviate from the reference item when asking the user to
  make choices. It's summed across both items; e.g., `max_dev == 5`
  allows 4 deviant criteria compared to 1 deviant criterion.
- `allowed_pairs_callback` is called on `allowed_pairs` for each
  iteration of the outermost loop."""

    criteria, alts, prefs = _setup(criteria, alts)
    assert 2 <= max_dev <= 2*len(criteria)
    if find_best:
        assert 1 <= find_best <= len(alts)

    async def get_pref(a, b):
        add_items(criteria, prefs, [a, b])
        if (rel := prefs.cmp(a, b)) is None:
            rel = asker(a, b)
            if inspect.isawaitable(rel):
                rel = await rel
            prefs.learn(a, b, rel)
        return rel

    def dev_from_ref(dev_criteria, vector):
        '''Return a vector in the item space that deviates from the
        best possible item on the given criteria with the given
        values.'''
        return tuple(
           vector[i] if i in dev_criteria else c[-1]
           for i, c in enumerate(criteria))

    def num_item(item):
        '''Represent criterion values as integers. This prevents us
        from behaving differently depending on the default sort order
        of the criterion values. (We're still sensitive to the order
        of criteria, though.)'''
        return tuple(criteria[i].index(v) for i, v in enumerate(item))

    for allowed_pairs in (l[::-1] for l in accumulate(
           [(big, small)] + ([] if big == small else [(small, big)])
           for deviation in range(1, max_dev)
           for big in range(deviation, deviation // 2, -1)
           for small in [deviation + 1 - big]
           if small <= len(criteria))):

        allowed_pairs_callback(allowed_pairs)

        to_try = set(choose2(sorted(alts, key = num_item)))

        while True:

            if find_best and len(prefs.maxes(alts)) >= find_best:
                return prefs

            # Don't ask about pairs we already know.
            to_try = {x for x in to_try if prefs.cmp(*x) == IC}

            if find_best:
                # Don't compare alternatives that can't be in the
                # requested `extreme` set.
                not_best = {x
                    for x in alts
                    if sum(prefs.cmp(x, a) == LT for a in alts) >=
                       find_best}
                to_try = {(a, b)
                    for a, b in to_try
                    if a not in not_best and b not in not_best}

            if not to_try:
                if any(prefs.cmp(a, b) == IC for a, b in choose2(alts)):
                    break
                return prefs

            a, b = max(to_try, key = lambda pair:
                (num_item(pair[0]), num_item(pair[1])))
            to_try.remove((a, b))

            cs = frozenset({ci
                for ci in range(len(criteria))
                if a[ci] != b[ci]})
            try:
                async def f(rel, cs1, cs2):
                    if not cs1:
                        raise Jump(rel)
                    for size1, size2 in allowed_pairs[:
                            # In the topmost call of `f`, force use of the
                            # new allowed_pairs, to save pointless iterations.
                            2 if len(cs1) == len(criteria) else None]:
                        if (len(cs1) - size1 < 0 or len(cs2) - size2 < 0 or
                                (len(cs1) - size1 == 0) != (len(cs2) - size2 == 0)):
                            continue
                        for c1 in combinations(sorted(cs1), size1):
                            for c2 in combinations(sorted(cs2), size2):
                                p = await get_pref(dev_from_ref(c1, a), dev_from_ref(c2, b))
                                if rel == EQ or p in (EQ, rel):
                                    await f(rel or p, cs1.difference(c1), cs2.difference(c2))
                await f(EQ, cs, cs)
            except Jump as j:
                prefs.learn(a, b, j.value)
            except Abort:
                return prefs

    return prefs

def _setup(criteria, alts = None, find_best = None):
    '''Some initial VDA logic put into its own function so it can be
    tested separately.'''

    criteria = tuple(map(tuple, criteria))
    assert len(criteria)
    assert all(
        len(c) > 0 and len(c) == len(set(c))
        for c in criteria)

    if alts is None:
        alts = tuple(product(*criteria))
    else:
        alts = tuple(map(tuple, alts))
        assert all(
            len(a) == len(criteria) and all(
                a[i] in criteria[i]
                for i in range(len(a)))
            for a in alts)
        assert len(alts) == len(set(alts))

    # Define the user's preferences as a preorder, with `a < b` if `b`
    # is preferred to `a`, and `a` and `b` incomparable if the user's
    # preference isn't yet known.
    prefs = PreorderedSet()
    add_items(criteria, prefs, alts)

    return criteria, alts, prefs

def add_items(criteria, prefs, items):
    # Enforce the assumption that on any single criterion, bigger
    # values are better.
    for x in set(items) - prefs.elements:
        prefs.add(x)
        for a in prefs.elements - {x}:
            cmps = [
                cmp(criteria[ci].index(x[ci]),
                    criteria[ci].index(a[ci]))
                for ci in range(len(criteria))]
            if not (LT in cmps and GT in cmps):
                # One item dominates the other. (We know that `cmps`
                # isn't all EQ because `x` and `a` are different.)
                prefs.learn(x, a, LT if LT in cmps else GT)



# Ensure that each module can accessed explicitly as `m`; e.g.,
# `artiruno.m.vda` will get the `vda` module. This is useful because
# `artiruno.vda` will be set to a function instead of the module.
import artiruno.util, artiruno.preorder, artiruno.vda, artiruno.interactive
class C: pass
m = C()
for s in ('util', 'preorder', 'vda', 'interactive'):
    setattr(m, s, getattr(artiruno, s))

# Only import `artiruno.web` on Brython, since its own imports will
# fail on CPython.
import platform
if platform.python_implementation() == 'Brython':
   from artiruno.web import initialize_web_interface

# Now do the ordinary imports.
from artiruno._version import __version__
from artiruno.util import *
from artiruno.preorder import (
    PreorderedSet, IC, EQ, LT, GT, ContradictionError)
from artiruno.vda import vda, avda



import itertools
from collections import Counter
from artiruno.util import choose2

IC, LT, EQ, GT = None, -1, 0, 1
  # The types of relations: incomparable, less than, equivalent, greater than

def invert_rel(rel):
    return (
        LT if rel == GT else
        GT if rel == LT else
        rel)

class ContradictionError(Exception):
    def __init__(self, k, was, claimed):
        super().__init__("{}: known to be {}, now claimed to be {}".format(k, was, claimed))

class PreorderedSet:
    '''A set equipped with a preorder. Elements can be added to
    the set, and the order can be updated to make previously
    incomparable elements comparable.'''

    def __init__(self, elements = (), relations = (), raw_relations = None):
        self.elements = set(elements)
        self.relations = raw_relations or {(a, b): IC
            for a, b in choose2(sorted(self.elements))}
        for a, b, rel in relations:
            self.learn(a, b, rel)

    def copy(self):
        return type(self)(self.elements.copy(),
            raw_relations = self.relations.copy())

    def add(self, x):
        if x in self.elements:
            return
        for a in self.elements:
            self.relations[(a, x) if a < x else (x, a)] = IC
        self.elements.add(x)

    def cmp(self, a, b):
        if a == b:
            assert a in self.elements and b in self.elements
            return EQ
        return self.relations[a, b] if a < b else invert_rel(self.relations[b, a])

    def extreme(self, n, among = None, bottom = False):
        rel = GT if bottom else LT
          # Work around https://github.com/brython-dev/brython/issues/1535
        return frozenset(x
            for x in among or self.elements
            for cmps in [Counter(
                self.cmp(x, a) for a in among or self.elements)]
            if cmps[IC] == 0 and cmps[rel] < n)

    def maxes(self, among = None):
        return self.extreme(1, among, bottom = False)
    def mins(self, among = None):
        return self.extreme(1, among, bottom = True)

    def _set(self, a, b, rel):
        'Return true if a change was made.'
        assert rel in (LT, EQ, GT)

        if a == b:
             assert a in self.elements
             if rel != EQ:
                 raise ContradictionError((a, b), EQ, rel)
             return False

        k = (a, b)
        if b < a:
           k, rel = (b, a), invert_rel(rel)

        if self.relations[k] == IC:
            self.relations[k] = rel
            return True
        elif self.relations[k] == rel:
            return False
        else:
            raise ContradictionError(k, self.relations[k], rel)

    def learn(self, a, b, rel):
        '''Update the ordering. Return a list of pairs that were updated.
        Raise ContradictionError if the new assertion isn't consistent
        with the preexisting non-IC relations.'''
        if not self._set(a, b, rel):
            return []
        # Use a modification of Warshall's algorithm to update the transitive
        # closure.
        # https://web.archive.org/web/2013/https://cs.winona.edu/lin/cs440/ch08-2.pdf
        changed = [(a, b) if a < b else (b, a)]
        for k, i, j in itertools.product((a, b), self.elements, self.elements):
            r1, r2 = self.cmp(i, k), self.cmp(k, j)
            if (r1 != IC and r2 == EQ) or (r1 == LT == r2):
                if self._set(i, j, r1):
                    changed.append((i, j) if i < j else (j, i))
        return changed

    def get_subset(self, elements):
        elements = frozenset(elements)
        assert elements.issubset(self.elements)
        return type(self)(elements, raw_relations = {(a, b): r
            for (a, b), r in self.relations.items()
            if {a, b}.issubset(elements)})

    def _summary(self):
        return " ".join(
            "{}{}{}".format(a, {EQ: "=", LT: "<"}[rel], b)
            for a, b, rel in sorted(
                (b, a, LT) if rel == GT else (a, b, rel)
                for (a, b), rel in self.relations.items()
                if rel != IC))

    def graph(self, namer = str):
        import subprocess
        import graphviz

        g = graphviz.Digraph()

        # Collect groups of equivalent items into nodes.
        nodes = []
        for a in sorted(self.elements):
            for node in nodes:
                if self.cmp(a, node[0]) == EQ:
                    node.append(a)
                    break
            else:
                nodes.append([a])

        # Add the nodes to the graph.
        def node_repr(node):
            return " / ".join(map(namer, sorted(node)))
        for node in nodes:
            g.node(node_repr(node))

        # Add edges.
        for a, b in itertools.product(nodes, nodes):
            if self.cmp(a[0], b[0]) == LT:
                g.edge(node_repr(b), node_repr(a))

        # Render to Graphviz source code, transform to the transitive
        # reduction with `tred`, and redigest the graph.
        return graphviz.Source(subprocess
            .Popen(["tred"], stdin = subprocess.PIPE, stdout = subprocess.PIPE)
            .communicate(g.source.encode('UTF-8'))[0].decode('UTF-8'))



import sys, json
from tempfile import mktemp
from artiruno.preorder import LT, EQ, GT
from artiruno.vda import vda, Abort
from artiruno._version import __version__

def interact(criterion_names, alts, alt_names, **kwargs):

    def asker(a, b):
        print('\nWhich do you prefer?')

        options = dict(a = GT, b = LT, e = EQ, q = Abort)
        for k in 'a', 'b':
            item = dict(a = a, b = b)[k]
            print(f'\n({k})' + (
                ' <<< ' + alt_names[alts.index(item)]
                    if alt_names and item in alts else ''))
            for name, value in zip(criterion_names, item):
                print(f'- {name}: {value}')
        print('\n(e) The two options are equally preferable')
        print('(q) Abort')

        print('')
        while True:
            try:
                inp = input('> ').strip().lower()
            except (EOFError, KeyboardInterrupt):
                print('Aborted.')
                raise Abort()
            try:
                v = options[inp]
            except KeyError:
                print('Invalid input.')
                continue
            if v is Abort:
                print('Aborted.')
                raise Abort()
            return v

    return vda(
        asker = asker,
        alts = alts,
        allowed_pairs_callback = apc,
        **kwargs)

def apc(allowed_pairs):
    if len(allowed_pairs) > 1:
        print("Allowed pairs now:", allowed_pairs[0])

def main():
    print('Artiruno', __version__)

    with open(sys.argv[1]) as o:
        scenario = json.load(o)

    alt_names = None
    if alts := scenario.get('alts'):
        if isinstance(alts, dict):
            alt_names, alts = zip(*alts.items())
        alts = tuple(tuple(a[c] for c in scenario['criteria'])
            for a in alts)

    prefs = interact(
        criterion_names = list(scenario['criteria']),
        criteria = list(scenario['criteria'].values()),
        alt_names = alt_names,
        alts = alts,
        find_best = scenario.get('find_best'),
        max_dev = 2 * len(scenario['criteria']))

    print("Best:", ", ".join(
        alt_names[alts.index(a)] if alt_names else str(a)
        for a in prefs.maxes(alts)))

    (prefs.get_subset(alts)
        .graph(namer = (lambda a: alt_names[alts.index(a)])
            if alt_names else str)
        .render(filename = mktemp(), format = 'png', view = True))



# This module is only run by Brython.

import browser
import browser.aio as aio
import browser.html as H
import javascript as JS
from artiruno.vda import avda
from artiruno.preorder import LT, GT, EQ

document, window = browser.document, browser.window

vda_running = False

class Quit(Exception): pass

def initialize_web_interface():
    with open('examples/jobs.json', 'r') as o:
        document['problem-definition'].value = o.read().strip()
    (document['start-button-parent'] <=
        H.BUTTON('Start decision-making', id = 'start-button'))
    document['start-button'].bind('click', restart_decision_making)

def restart_decision_making(_):
    aio.run(_restart_decision_making())

async def _restart_decision_making():
    global vda_running

    # Terminate any current VDA and clear the log.
    if vda_running:
        # Kill the current VDA job.
        signal('choice', 'quit')
        # Wait till it's dead before continuing.
        await get_signal('quit_done')
    document['dm'].textContent = ''

    # Reword "start" to "restart".
    document['start-button'].textContent = 'Restart decision making'

    # Parse the problem definition.
    try:
        scenario = JS.JSON.parse(document['problem-definition'].value)
    except Exception as e:
        # There's no more specific exception type to use at the Python
        # level.
        document['dm'].textContent = str(e)
        return

    # Set up the alts.
    alt_names = None
    if alts := scenario.get('alts'):
        if isinstance(alts, dict):
            alt_names, alts = zip(*alts.items())
        alts = tuple(tuple(a[c] for c in scenario['criteria'])
            for a in alts)

    # Begin VDA.
    try:
        vda_running = True
        prefs = await interact(
            criterion_names = list(scenario['criteria']),
            criteria = list(scenario['criteria'].values()),
            alt_names = alt_names,
            alts = alts,
            find_best = scenario.get('find_best'),
            max_dev = 2 * len(scenario['criteria']))
        document['dm'] <= H.P('Best: ' + ", ".join(
            alt_names[alts.index(a)] if alt_names else str(a)
            for a in prefs.maxes(alts)))
    except Quit:
        signal('quit_done')
    finally:
        vda_running = False

n_questions = 0

async def interact(criterion_names, alts, alt_names, **kwargs):
    global n_questions
    n_questions = 0

    async def asker(a, b):
        # Present the choices as a list with buttons.

        global n_questions
        n_questions += 1

        buttons = {}
        def button(the_id, text):
            buttons[the_id] = lambda _: signal('choice', the_id)
            return H.BUTTON(text, id = the_id)

        def display_item(item):
            # Display the item as a list with one criterion and value
            # per list item. Highlight the criteria value that differ
            # between the two options.
            return H.UL([
                 H.LI([name + ': ',
                     (value if a[i] == b[i] else H.STRONG(value))])
                 for i, (name, value)
                 in enumerate(zip(criterion_names, item))])

        document['dm'] <= H.DIV(
            [H.P('Q{}: Which would you prefer?'.format(n_questions)), H.UL([
                H.LI([button('option_a', 'Option A'),
                    display_item(a)]),
                H.LI([button('option_b', 'Option B'),
                    display_item(b)]),
                H.LI([button('equal', 'Equal'),
                    'The two options are equally preferable'])])],
            Class = 'query')

        for button, callback in buttons.items():
            document[button].bind('click', callback)

        # Wait for the user to click a button.
        choice = await get_signal('choice')
        if choice == 'quit':
            raise Quit()

        # Replace the buttons with indicators of the user's decision.
        for button in buttons:
            document[button].replaceWith(H.SPAN(
               'your choice' if button == choice else 'not chosen',
               Class = 'chosen' if button == choice else 'not-chosen'))

        # Return the choice.
        return dict(option_a = GT, option_b = LT, equal = EQ)[choice]

    return await avda(
        asker = asker,
        alts = alts,
        **kwargs)

def signal(event_type, detail = None):
    document.dispatchEvent(window.CustomEvent.new(
        event_type, dict(detail = detail)))

async def get_signal(event_type):
    return (await aio.event(document, event_type)).detail
